import logging
from abc import ABC, abstractmethod
from asyncio import sleep
from datetime import datetime, timedelta
from json.decoder import JSONDecodeError
from math import ceil, floor
from time import monotonic
from typing import Annotated, Any, ClassVar, Dict, List, Optional

from httpx import AsyncClient, Response
from pydantic import BaseModel, BeforeValidator, ConfigDict
from starlette import status
from tenacity import retry
from tenacity.retry import retry_if_exception, retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from spylib.constants import (
    API_CALL_NUMBER_RETRY_ATTEMPTS,
    MAX_COST_EXCEEDED_ERROR_CODE,
    OPERATION_NAME_REQUIRED_ERROR_MESSAGE,
    THROTTLED_ERROR_CODE,
    WRONG_OPERATION_NAME_ERROR_MESSAGE,
)
from spylib.exceptions import (
    ShopifyCallInvalidError,
    ShopifyError,
    ShopifyExceedingMaxCostError,
    ShopifyGQLError,
    ShopifyIntermittentError,
    ShopifyInvalidResponseBody,
    ShopifyThrottledError,
    not_our_fault,
)
from spylib.utils.misc import TimedResult, elapsed_time, parse_scope
from spylib.utils.rest import Request


class Token(ABC, BaseModel):
    """Abstract class for token objects.

    This should never be extended, as you should either be
    using the OfflineTokenABC or the OnlineTokenABC.
    """

    store_name: str
    scope: Annotated[List[str], BeforeValidator(parse_scope)] = []
    access_token: Optional[str] = None
    access_token_invalid: bool = False

    api_version: ClassVar[Optional[str]] = None

    rest_bucket_max: int = 80
    rest_bucket: int = rest_bucket_max
    rest_leak_rate: int = 4

    graphql_bucket_max: int = 1000
    graphql_bucket: int = graphql_bucket_max
    graphql_leak_rate: int = 50

    updated_at: float = monotonic()

    client: ClassVar[AsyncClient] = AsyncClient()

    @property
    def oauth_url(self) -> str:
        return f'https://{self.store_name}.myshopify.com/admin/oauth/access_token'

    @property
    def api_url(self) -> str:
        if not self.api_version:
            return f'https://{self.store_name}.myshopify.com/admin'
        return f'https://{self.store_name}.myshopify.com/admin/api/{self.api_version}'

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # Methods for querying the store

    async def __await_rest_bucket_refill(self):
        self.__fill_rest_bucket()
        while self.rest_bucket <= 1:
            await sleep(1)
            self.__fill_rest_bucket()
        self.rest_bucket -= 1

    def __fill_rest_bucket(self):
        now = monotonic()
        time_since_update = now - self.updated_at
        new_tokens = floor(time_since_update * self.rest_leak_rate)
        if new_tokens > 1:
            self.rest_bucket = min(self.rest_bucket + new_tokens, self.rest_bucket_max)
            self.updated_at = now

    async def __handle_error(self, debug: str, endpoint: str, response: Response):
        """Handle any error that occured when calling Shopify.

        If the response has a valid json then return it too.
        """
        msg = (
            f'ERROR in store {self.store_name}: {debug}\n'
            f'API response code: {response.status_code}\n'
            f'API endpoint: {endpoint}\n'
        )
        try:
            jresp = response.json()
        except Exception:
            pass
        else:
            msg += f'API response json: {jresp}\n'

        if 400 <= response.status_code < 500:
            # This appears to be our fault
            raise ShopifyCallInvalidError(msg)

        raise ShopifyError(msg)

    @retry(
        reraise=True,
        wait=wait_random(min=1, max=2),
        stop=stop_after_attempt(API_CALL_NUMBER_RETRY_ATTEMPTS),
        retry=retry_if_exception(not_our_fault),
    )
    async def execute_rest(
        self,
        request: Request,
        endpoint: str,
        json: Optional[Dict[str, Any]] = None,
        debug: str = '',
    ) -> Dict[str, Any]:
        while True:
            await self.__await_rest_bucket_refill()

            if not self.access_token:
                raise ValueError('You have not initialized the token for this store. ')

            response = await self.client.request(
                method=request.method.value,
                url=f'{self.api_url}{endpoint}',
                headers={'X-Shopify-Access-Token': self.access_token},
                json=json,
            )
            if response.status_code == status.HTTP_429_TOO_MANY_REQUESTS:
                # We hit the limit, we are out of tokens
                self.rest_bucket = 0
                continue
            elif 400 <= response.status_code or response.status_code != request.good_status:
                # All errors are handled here
                await self.__handle_error(debug=debug, endpoint=endpoint, response=response)
            else:
                jresp = response.json()
                # Recalculate the rate to be sure we have the right one.
                calllimit = response.headers['X-Shopify-Shop-Api-Call-Limit']
                self.rest_bucket_max = int(calllimit.split('/')[1])
                # In Shopify the bucket is emptied after 20 seconds
                # regardless of the bucket size.
                self.rest_leak_rate = int(self.rest_bucket_max / 20)

            return jresp

    @retry(
        reraise=True,
        stop=stop_after_attempt(API_CALL_NUMBER_RETRY_ATTEMPTS),
        retry=retry_if_exception_type(
            (ShopifyThrottledError, ShopifyInvalidResponseBody, ShopifyIntermittentError)
        ),
    )
    async def execute_gql(
        self,
        query: str,
        variables: Dict[str, Any] = {},
        operation_name: Optional[str] = None,
        suppress_errors: bool = False,
    ) -> Dict[str, Any]:
        if not self.access_token:
            raise ValueError('Token Undefined')

        url = f'{self.api_url}/graphql.json'

        headers = {
            'Content-type': 'application/json',
            'X-Shopify-Access-Token': self.access_token,
        }

        body = {'query': query, 'variables': variables, 'operationName': operation_name}

        resp = await self.client.post(
            url=url,
            json=body,
            headers=headers,
        )

        # Handle any response that is not 200, which will return with error message
        # https://shopify.dev/api/admin-graphql#status_and_error_codes
        if resp.status_code >= 500:
            raise ShopifyIntermittentError(
                f'The Shopify API returned an intermittent error: {resp.status_code}.'
            )

        if resp.status_code != 200:
            try:
                jsondata = resp.json()
                error_msg = f'{resp.status_code}. {jsondata["errors"]}'
            except JSONDecodeError:
                error_msg = f'{resp.status_code}.'

            raise ShopifyGQLError(f'GQL query failed, status code: {error_msg}')

        try:
            jsondata = resp.json()
        except JSONDecodeError as exc:
            raise ShopifyInvalidResponseBody from exc

        if type(jsondata) is not dict:
            raise ValueError('JSON data is not a dictionary')
        if 'Invalid API key or access token' in jsondata.get('errors', ''):
            self.access_token_invalid = True
            logging.warning(
                f'Store {self.store_name}: The Shopify API token is invalid. '
                'Flag the access token as invalid.'
            )
            raise ConnectionRefusedError

        if 'data' not in jsondata and 'errors' in jsondata:
            errorlist = '\n'.join(
                [err['message'] for err in jsondata['errors'] if 'message' in err]
            )
            error_code_list = '\n'.join(
                [
                    err['extensions']['code']
                    for err in jsondata['errors']
                    if 'extensions' in err and 'code' in err['extensions']
                ]
            )
            if MAX_COST_EXCEEDED_ERROR_CODE in error_code_list:
                raise ShopifyExceedingMaxCostError(
                    f'Store {self.store_name}: This query was rejected by the Shopify'
                    f' API, and will never run as written, as the query cost'
                    f' is larger than the max possible query size (>{self.graphql_bucket_max})'
                    ' for Shopify.'
                )
            elif THROTTLED_ERROR_CODE in error_code_list:  # This should be the last condition
                query_cost = jsondata['extensions']['cost']['requestedQueryCost']
                available = jsondata['extensions']['cost']['throttleStatus']['currentlyAvailable']
                rate = jsondata['extensions']['cost']['throttleStatus']['restoreRate']
                sleep_time = ceil((query_cost - available) / rate)
                await sleep(sleep_time)
                raise ShopifyThrottledError
            elif OPERATION_NAME_REQUIRED_ERROR_MESSAGE in errorlist:
                raise ShopifyCallInvalidError(
                    f'Store {self.store_name}: Operation name was required for this query.'
                    'This likely means you have multiple queries within one call '
                    'and you must specify which to run.'
                )
            elif WRONG_OPERATION_NAME_ERROR_MESSAGE.format(operation_name) in errorlist:
                raise ShopifyCallInvalidError(
                    f'Store {self.store_name}: Operation name {operation_name}'
                    'does not exist in the query.'
                )
            else:
                raise ValueError(f'GraphQL query is incorrect:\n{errorlist}')

        if not suppress_errors and len(jsondata.get('errors', [])) >= 1:
            raise ShopifyGQLError(jsondata)

        return jsondata['data']

    @elapsed_time(data_type=TimedResult)
    async def test_connection(self) -> bool:
        """Test the connection to the Shopify Admin APIs."""
        try:
            await self.execute_gql(
                query='query { shop { name } }',
            )
            logging.info('Shopify API connection is OK')
        except Exception as e:
            logging.exception(e)
            return False
        return True


class OfflineTokenABC(Token, ABC):
    """Offline tokens are used for long term access, and do not have a set expiry."""

    @abstractmethod
    async def save(self):
        pass

    @classmethod
    @abstractmethod
    async def load(cls, store_name: str):
        pass


class OnlineTokenABC(Token, ABC):
    """Online tokens are used to implement applications authenticated with a specific user's credentials.

    These extend on the original token, adding in a user, its scope and an expiry time.
    """

    associated_user_id: int
    expires_in: int = 0
    expires_at: datetime = datetime.now() + timedelta(days=0, seconds=expires_in)

    @abstractmethod
    async def save(self):
        """This method handles saving the token.

        By default this does nothing, therefore the developer should override this.
        """

    @classmethod
    @abstractmethod
    async def load(cls, store_name: str, associated_user: str):
        """This method handles loading the token.

        By default this does nothing, therefore the developer should override this.
        """


class PrivateTokenABC(Token, ABC):
    """Private token implementation, when we are pulling this from the config file.

    Therefore we do not need the save function for the token class as there is
    no calls to the OAuth endpoints for shopify.
    """

    @classmethod
    @abstractmethod
    async def load(cls, store_name: str):
        """This method handles loading the token.

        By default this does nothing, therefore the developer should override this.
        """
        pass



from __future__ import annotations

from typing import Any, Dict, Optional
from urllib.parse import urlparse

from jwt import decode
from pydantic import model_validator
from pydantic.main import BaseModel
from pydantic.networks import HttpUrl

from .utils.domain import store_domain

REQUIRED_FIELDS = ['iss', 'dest', 'sub', 'jti', 'sid']
ALGORITHM = 'HS256'
PREFIX = 'Bearer '
LEEWAY_SECONDS = 10


class TokenValidationError(Exception):
    pass


class InvalidIssuerError(TokenValidationError):
    pass


class MismatchedHostError(TokenValidationError):
    pass


class TokenAuthenticationError(TokenValidationError):
    pass


class SessionToken(BaseModel):
    """Session tokens are derived from the authorization header from Shopify.

    This performs the set of validations as defined by shopify
    https://shopify.dev/apps/auth/session-tokens/authenticate-an-embedded-app-using-session-tokens#obtain-session-details-manually
    """

    iss: HttpUrl
    dest: HttpUrl
    aud: Optional[str] = None
    sub: str
    exp: Optional[float] = None
    nbf: Optional[float] = None
    iat: Optional[float] = None
    jti: str
    sid: str

    @model_validator(mode='before')
    @classmethod
    def equal_iss_and_dest(cls, values: Dict[str, Any]):
        domain = cls.__url_to_base(values.get('iss'))
        try:
            store_domain(domain)
        except ValueError as e:
            raise InvalidIssuerError(f'The domain {domain} is not a valid issuer.') from e

        iss = cls.__url_to_base(values.get('iss'))
        dest = cls.__url_to_base(values.get('dest'))
        if iss != dest:
            raise MismatchedHostError(f'The issuer {iss} does not match the destination {dest}')

        return values

    @classmethod
    def from_header(
        cls,
        authorization_header: str,
        api_key: str,
        secret: str,
    ) -> SessionToken:
        # Take the authorization headers and unload them
        if not authorization_header.startswith(PREFIX):
            raise TokenAuthenticationError(
                'The authorization header does not contain a Bearer token.'
            )

        token = authorization_header[len(PREFIX) :]

        payload = decode(
            token,
            secret,
            audience=api_key,
            algorithms=[ALGORITHM],
            # AppBridge frequently sends future `nbf`, and it causes `ImmatureSignatureError`.
            # Accept few seconds clock skew to avoid this error.
            leeway=LEEWAY_SECONDS,
            options={'require': REQUIRED_FIELDS},
        )

        # Verify enough fields specified and perform validation checks
        return cls.model_validate(payload)

    @staticmethod
    def __url_to_base(url):
        return '{uri.scheme}://{uri.netloc}'.format(uri=urlparse(url))



from datetime import datetime, timezone
from functools import wraps
from time import perf_counter
from typing import Any, List, Type

from pydantic import BaseModel

from .shortuuid import random


def now_epoch() -> int:
    return int(datetime.now(timezone.utc).timestamp())


def get_unique_id() -> str:
    return random(length=10)


def parse_scope(v: Any) -> List[str]:
    if isinstance(v, str):
        return v.split(',')
    return v


class ElapsedTime(BaseModel):
    start: float
    end: float

    @property
    def seconds(self) -> float:
        return self.end - self.start

    @property
    def milliseconds(self) -> float:
        return 1000 * (self.end - self.start)


class TimedResult(BaseModel):
    result: Any
    elapsed_time: ElapsedTime


def elapsed_time(data_type: Type[TimedResult]):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            start = perf_counter()
            result = await func(*args, **kwargs)
            end = perf_counter()

            return data_type(result=result, elapsed_time=ElapsedTime(start=start, end=end))

        return wrapper

    return decorator



from .domain import domain_to_storename, store_domain
from .httpclient import HTTPClient
from .jwtoken import JWTBaseModel
from .misc import TimedResult, elapsed_time, get_unique_id, now_epoch
from .rest import DELETE, GET, POST, PUT, Method

__all__ = [
    'now_epoch',
    'get_unique_id',
    'JWTBaseModel',
    'HTTPClient',
    'domain_to_storename',
    'store_domain',
    'Method',
    'GET',
    'POST',
    'PUT',
    'DELETE',
    'TimedResult',
    'elapsed_time',
]



class ShopifyError(Exception):
    """Exception to identify any Shopify error."""

    pass


class ShopifyGQLError(Exception):
    """Exception to identify any Shopify admin graphql error."""

    pass


class ShopifyIntermittentError(Exception):
    """Exception to identify any Shopify admin API Intermittent error."""

    pass


class ShopifyGQLUserError(Exception):
    """Exception to identify any Shopify admin graphql error caused by the user mistake."""

    pass


class ShopifyInvalidResponseBody(ShopifyError):
    """Exception to identify when Shopify returned body is not the expected type, i.e. not a JSON.

    These seem extremely rare and should be retried because they are intermittent and most likely
    and internal error in Shopify that wasn't handled properly.
    """

    pass


class ShopifyCallInvalidError(ShopifyError):
    """Exception to identify errors that our due to bad data sent by us.

    These should not be retried
    """

    pass


class ShopifyThrottledError(ShopifyError):
    """Exception to identify errors that are due to rate limit control."""

    pass


class ShopifyExceedingMaxCostError(ShopifyError):
    """Exception to identify errors that are due to queries exceeding the max query size."""

    pass


class FastAPIImportError(ImportError):
    """Exception to identify errors when spylip.oauth is accessed without fastapi installed."""

    pass


def not_our_fault(exception: BaseException) -> bool:
    """Simple function to identify invalid Shopify calls, i.e. our mistake.

    Probably the only way to make sure we retry for any other exception but those.
    """
    return not isinstance(exception, ShopifyCallInvalidError)



from typing import Any, List, Tuple
from urllib.parse import parse_qsl

from spylib.hmac import validate as validate_hmac

from ..utils import domain_to_storename, now_epoch
from .tokens import OAuthJWT


def validate_callback(shop: str, timestamp: int, query_string: Any, api_secret_key: str) -> None:
    q_str = query_string.decode('utf-8')
    # 1) Check that the shop is a valid Shopify URL
    domain_to_storename(shop)

    # 2) Check the timestamp. Must not be more than 5min old
    if now_epoch() - timestamp > 300:
        raise ValueError('Timestamp is too old')

    # 3) Check the hmac
    # Extract HMAC
    args = parse_qsl(q_str)
    validate_callback_args(args=args, api_secret_key=api_secret_key)


def validate_callback_args(args: List[Tuple[str, str]], api_secret_key: str) -> None:
    # We assume here that the arguments were validated prior to calling
    # this function.
    hmac_arg = [arg[1] for arg in args if arg[0] == 'hmac'][0]
    message = '&'.join([f'{arg[0]}={arg[1]}' for arg in args if arg[0] != 'hmac'])
    # Check HMAC
    validate_hmac(secret=api_secret_key, sent_hmac=hmac_arg, message=message)


def validate_oauthjwt(token: str, shop: str, jwt_key: str) -> OAuthJWT:
    oauthjwt = OAuthJWT.decode_token(token=token, key=jwt_key)

    storename = domain_to_storename(shop)
    if oauthjwt.storename != storename:
        raise ValueError("Token storename and query shop don't match")

    return oauthjwt



from base64 import b64encode
from hashlib import sha256
from hmac import compare_digest, new


def calculate_from_message(secret: str, message: str, use_base64: bool = False) -> str:
    hmac_hash = new(secret.encode('utf-8'), message.encode('utf-8'), sha256)
    if use_base64:
        # TODO fix bytes / str union issue
        return b64encode(hmac_hash.digest()).decode('utf-8')

    return hmac_hash.hexdigest()


def calculate_from_components(
    datetime: str, path: str, query_string: str, body: str, secret: str, use_base64: bool = False
) -> str:
    if query_string != '':
        path = path + '?' + query_string
    message = path + datetime + body
    return calculate_from_message(secret=secret, message=message, use_base64=use_base64)


def validate(secret: str, sent_hmac: str, message: str, use_base64: bool = False):
    hmac_calculated = calculate_from_message(secret=secret, message=message, use_base64=use_base64)

    if not compare_digest(sent_hmac, hmac_calculated):
        raise ValueError('HMAC verification failed')



from spylib.exceptions import FastAPIImportError

try:
    from .authentication import (
        WebhookHMACHeader,
        authenticate_webhook_hmac,
        webhook_hmac,
    )
except ImportError as e:
    raise FastAPIImportError(
        'The fastapi_extensions require `fastapi` which is not installed. '
        'Run `pip install spylib[fastapi]` to be able to use it.'
    ) from e

__all__ = [
    'webhook_hmac',
    'authenticate_webhook_hmac',
    'WebhookHMACHeader',
]



from .exchange_token import (
    exchange_offline_token,
    exchange_online_token,
    exchange_token,
)
from .models import OfflineTokenModel, OnlineTokenModel
from .signature_validation import validate_signed_query_string

__all__ = [
    'exchange_token',
    'exchange_offline_token',
    'exchange_online_token',
    'OfflineTokenModel',
    'OnlineTokenModel',
    'validate_signed_query_string',
]



# Shopify GraphQL error Codes
# https://shopify.dev/api/admin-graphql#status_and_error_codes
THROTTLED_ERROR_CODE = 'THROTTLED'
MAX_COST_EXCEEDED_ERROR_CODE = 'MAX_COST_EXCEEDED'

# Error Messages
OPERATION_NAME_REQUIRED_ERROR_MESSAGE = 'An operation name is required'
WRONG_OPERATION_NAME_ERROR_MESSAGE = (
    'No operation named "{}"'  # Required as they don't stay consistant
)

UTF8ENCODING = 'utf-8'
API_CALL_NUMBER_RETRY_ATTEMPTS = 5



from typing import List

from ..utils import domain_to_storename, get_unique_id
from .tokens import OAuthJWT


def oauth_init_url(
    domain: str,
    requested_scopes: List[str],
    callback_domain: str,
    callback_path: str,
    path_prefix: str,
    is_login: bool,
    jwt_key: str,
    api_key: str,
) -> str:
    """Create the URL and the parameters needed to start the oauth process to install an app or to log a user in.

    Args
    ----------
    domain: Domain of the shopify store in the format "{storesubdomain}.myshopify.com"
    requested_scopes: List of scopes accessible to the app once installed.
        See https://shopify.dev/docs/admin-api/access-scopes
    callback_domain: Public domain Shopify will connect to during the oauth process
    is_login: Specify if the oauth is to install the app or a user logging in

    Returns
    -------
    URL with all needed parameters to trigger the oauth process
    """
    scopes = ','.join(requested_scopes)
    redirect_uri = f'https://{callback_domain}{path_prefix}{callback_path}'
    oauthjwt = OAuthJWT(
        is_login=is_login, storename=domain_to_storename(domain), nonce=get_unique_id()
    )
    oauth_token = oauthjwt.encode_token(key=jwt_key)
    access_mode = 'per-user' if is_login else ''

    return (
        f'https://{domain}/admin/oauth/authorize?client_id={api_key}&'
        f'scope={scopes}&redirect_uri={redirect_uri}&state={oauth_token}&'
        f'grant_options[]={access_mode}'
    )


def app_redirect(
    store_domain: str,
    app_domain: str,
    app_api_key: str,
) -> str:
    return f'https://{store_domain}/admin/apps/{app_api_key}'



import datetime
import json
from base64 import urlsafe_b64encode
from typing import Any, Dict

from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA256
from Crypto.Random import get_random_bytes


def generate_token(secret: str, customer_data: Dict[str, Any]) -> bytes:
    key = SHA256.new(secret.encode('utf-8')).digest()
    encryption_key = key[0:16]
    signature_key = key[16:32]

    if 'email' not in customer_data:
        raise ValueError('Missing email in customer data')

    customer_data['created_at'] = datetime.datetime.utcnow().isoformat()
    cypher_text = _encrypt(encryption_key, json.dumps(customer_data))
    return urlsafe_b64encode(cypher_text + _sign(signature_key, cypher_text))


def generate_url(secret: str, customer_data: Dict[str, Any], store_url) -> str:
    token = generate_token(secret, customer_data).decode('utf-8')
    return f'{store_url}/account/login/multipass/{token}'


def _encrypt(encryption_key, plain_text) -> bytes:
    plain_text = _pad(plain_text)
    iv = get_random_bytes(AES.block_size)
    cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
    return iv + cipher.encrypt(str.encode(plain_text))


def _sign(signature_key, cypher_text):
    return HMAC.new(signature_key, cypher_text, SHA256).digest()


def _pad(s):
    return s + (AES.block_size - len(s) % AES.block_size) * chr(
        AES.block_size - len(s) % AES.block_size
    )
